/**************************************************************************\
*  HighLnk 0.2                            http://www.thpinfo.com/highlnk/  *
*      This source code is released under the terms in the LICENSE file.   *
*  Copyright (c) 2004 Thomas Perl <perl.thomas@aon.at>                     *
\**************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <linux/limits.h>

#include "md5.h"
#include "highlnk.h"
#include "finder.h"

void process_data( const char *fn, const struct stat *st);
void free_dir_source( struct dir_source *ds);
void free_file_source( struct file_source *fs);
void free_file_target( struct file_target *ft);

struct file_source *find_file_source( uint8 *digest);

struct file_source *fs = NULL, *fs_end = NULL;
struct dir_source *ds = NULL, *ds_end = NULL;
char working_directory[PATH_MAX], target_directory[PATH_MAX], source_path[PATH_MAX], target_path[PATH_MAX];
unsigned long int saved_space = 0, total_space = 0, files = 0, processed = 0;

int main( int argc, char* argv[])
{
  struct stat s[2], x;
  struct file_source *fspos;
  struct dir_source *dspos;
  struct file_target *ftpos;

  fprintf( stdout, "\n  %s %s -- Copyright (c) 2004 Thomas Perl <perl.thomas@aon.at>\n\n", basename( argv[0]), VERSION);

  getcwd( working_directory, PATH_MAX);

  if( argc != 2)
  { 
    print_usage_message();
    return 1;
  }

  strncpy( target_directory, argv[1], PATH_MAX);
  if( target_directory[strlen( target_directory)-1] == '/')
    target_directory[strlen( target_directory)-1] = '\0';

  stat( working_directory, &s[0]);

  if( 0 != stat( target_directory, &s[1]))
  {
    if( 0 != mkdir( target_directory, 0755))
    {
      fprintf( stderr, "%s cannot be created.\n", target_directory);
      return 2;
    }
    else
    {
#ifdef DEBUG
      fprintf( stderr, "Created directory: %s\n", target_directory);
#endif
      stat( target_directory, &s[1]);
    }
  }

  if( s[0].st_dev != s[1].st_dev)
  {
    fprintf( stderr, "%s and %s are not on the same device.\n", target_directory, working_directory);
    return 2;
  }


  fprintf( stdout, "Scanning %s for files\n", working_directory);
  findFiles( working_directory, &process_data);
  fprintf( stdout, "\n\n");
  
  fprintf( stdout, "Writing directory structure to %s\n", target_directory);
  dspos = ds;
  while( dspos != NULL)
  {
    strcpy( target_path, target_directory);
    strcat( target_path, dspos->dirname);
    strcpy( source_path, working_directory);
    strcat( source_path, dspos->dirname);

    stat( source_path, &x);

    if( s[1].st_ino != x.st_ino)
    {
#ifdef DEBUG
      fprintf( stderr, "Creating directory: %s\n", target_path);
#endif
      mkdir( target_path, x.st_mode & PERM_MASK);
      chmod( target_path, x.st_mode & PERM_MASK);
    }
#ifdef DEBUG
    else
    {
      fprintf( stderr, "Skipping target directory in source tree: %s\n", source_path);
    }
#endif

    dspos = dspos->next;
  }
  fprintf( stdout, "Directory structure written.\n\n");
  
  fprintf( stdout, "Writing file structure to %s\n", target_directory);
  fspos = fs;
  while( fspos != NULL)
  {
#ifdef DEBUG
    fprintf( stderr, "Creating links from file: %s\n", source_path);
#endif
    strcpy( source_path, working_directory);
    strcat( source_path, fspos->filename);
    
    ftpos = fspos->targets;
    while( ftpos != NULL)
    {
#ifdef DEBUG
      fprintf( stderr, "Creating link to file: %s\n", target_path);
#endif
      strcpy( target_path, target_directory);
      strcat( target_path, ftpos->filename);
      link( source_path, target_path);

      processed++;
      fprintf( stdout, "\r%.0f%% done (%d/%d files)", (processed*100.0)/(files*1.0), processed, files);

      ftpos = ftpos->next;
    }


    fspos = fspos->next;
  }
  fprintf( stdout, "\nFile structure written.\n\n");

  free_file_source( fs);
  free_dir_source( ds);

  fprintf( stdout, "Operation finished. Target directory is %d kb smaller.\n\n", saved_space/1024);

  return 0;
}

void process_data( const char *fn, const struct stat *st)
{
  static struct file_target *nft;
  static struct file_source *fspos, *nfs;
  static struct dir_source *nds;
  static uint8 digest[16];

  fprintf( stdout, "\r%d kb / %d files (%.2f%% redundancy)  ", total_space/1024, files, (saved_space*100.0)/(total_space*1.0));

#ifdef DEBUG
  fprintf( stderr, "Processing: %s\n", fn);
#endif

  if( S_ISDIR( st->st_mode))
  {
    nds = (struct dir_source*)malloc( sizeof( struct dir_source));
    nds->dirname = (char*)malloc( strlen( (char*)(fn+strlen( working_directory)))+1);
    strcpy( nds->dirname, (char*)(fn+strlen( working_directory)));
    nds->next = NULL;

    if( ds == NULL)
    {
      ds = nds;
      ds_end = nds;
    }
    else
    {
      ds_end->next = nds;
      ds_end = nds;
    }

#ifdef DEBUG
    fprintf( stderr, "Adding directory: %s\n", nds->dirname);
#endif

    return;
  }

  if( S_ISREG( st->st_mode))
  {
    calc_md5( fn, digest);

    fspos = find_file_source( digest);
    
    if( fspos == NULL)
    {
      nfs = (struct file_source*)malloc( sizeof( struct file_source));
      nfs->filename = (char*)malloc( strlen( (char*)(fn+strlen( working_directory)))+1);
      strcpy( nfs->filename, (char*)(fn+strlen(working_directory)));
      nfs->digest = (uint8*)malloc( sizeof( uint8) * 16);
      memcpy( nfs->digest, digest, sizeof( uint8) * 16);
      nfs->targets = NULL;
      nfs->targets_end = NULL;
      nfs->next = NULL;

      if( fs == NULL)
      {
        fs = nfs;
        fs_end = nfs;
      }
      else
      {
        fs_end->next = nfs;
        fs_end = nfs;
      }

      fspos = nfs;
    }
    else
    {
      saved_space += st->st_size;
#ifdef DEBUG
      fprintf( stderr, "Same files: %s %s\n", fspos->filename, (char*)(fn+strlen( working_directory)));
#endif
    }

    nft = (struct file_target*)malloc( sizeof( struct file_target));
    nft->filename = (char*)malloc( strlen( (char*)(fn+strlen( working_directory)))+1);
    strcpy( nft->filename, (char*)(fn+strlen(working_directory)));
    nft->next = NULL;
    total_space += st->st_size;
    files++;

    if( fspos->targets == NULL)
    {
      fspos->targets = nft;
      fspos->targets_end = nft;
    }
    else
    {
      fspos->targets_end->next = nft;
      fspos->targets_end = nft;
    }

#ifdef DEBUG
    fprintf( stderr, "Adding file: %s\n", nft->filename);
#endif

    return;
  }
}

void free_dir_source( struct dir_source *ds)
{
  struct dir_source *next, *pos = ds;

  while( pos != NULL)
  {
#ifdef DEBUG
    fprintf( stderr, "Freeing directory: %s\n", pos->dirname);
#endif
    next = pos->next;
    free( pos->dirname);
    free( pos);
    pos = next;
  }
}

void free_file_source( struct file_source *fs)
{
  struct file_source *next, *pos = fs;

  while( pos != NULL)
  {
#ifdef DEBUG
    fprintf( stderr, "Freeing file source: %s\n", pos->filename);
#endif
    next = pos->next;
    free( pos->filename);
    free( pos->digest);
    free_file_target( pos->targets);
    free( pos);
    pos = next;
  }
}

void free_file_target( struct file_target *ft)
{
  struct file_target *next, *pos = ft;

  while( pos != NULL)
  {
#ifdef DEBUG
    fprintf( stderr, "Freeing file target: %s\n", pos->filename);
#endif
    next = pos->next;
    free( pos->filename);
    free( pos);
    pos = next;
  }
}

struct file_source *find_file_source( uint8 *digest)
{
  struct file_source *pos = fs;

  while( pos != NULL)
  {
    if( memcmp( pos->digest, digest, sizeof( uint8) * 16) == 0)
      return pos;

    pos = pos->next;
  }

  return NULL;
}

